/**
 * Author: dnj
 * Date: August 29, 2008
 * 6.005 Elements of Software Construction
 * (c) 2008, MIT and Daniel Jackson
 */

package pianoapplet;
import java.awt.*;
import java.awt.event.*;
import javax.sound.midi.MidiUnavailableException;
import java.applet.*;
import midi.*;
import music.*;
import java.util.LinkedList;

/**
 * A skeletal applet that shows how to bind methods to key events
 */
public class PianoApplet extends Applet {
	
	private boolean keyLocked[];
	private Instrument instr;
	private boolean record;
	private long record_position;
	private LinkedList recording;
	
	private class MusicAction {
		public long delay;
		public char note; 
		public boolean start;
		
		public MusicAction(char n, long d, boolean s)
		{
			start = s;
			delay = d;
			note = n;		
		}
	}
	
	public void init() {
		final Midi midi;
		keyLocked = new boolean[15];
		for(int i=0; i<15; i++)
		{
			keyLocked[i] = false;
		}
		instr = Instrument.PIANO;

		try {
			midi = new Midi();
		} catch (MidiUnavailableException e1) {
			e1.printStackTrace();
			return;
		}
		
		setBackground(Color.black);
		
		// this is a standard pattern for associating method calls with GUI events
		// the call to the constructor of KeyAdapter creates an object of an
		// anonymous subclass of KeyAdapter, whose keyPressed method is called
		// when a key is pressed in the GUI
		addKeyListener(new KeyAdapter() {
			public void keyPressed(KeyEvent e) {
				char key = (char) e.getKeyCode();
				if(keyToInt(key) >= 0 && !keyLocked[keyToInt(key)])
				{
					keyLocked[keyToInt(key)] = true;
					if(keyToPitch(key)!= null)
					{
						midi.beginNote(keyToPitch(key).toMidiFrequency(), instr);
						if(record)
						{
							recording.add(new MusicAction('w', System.currentTimeMillis()-record_position, false));
							recording.add(new MusicAction(key, 0, true));
							record_position = System.currentTimeMillis();
						}
					}
					else if(key == 'I')
					{
						//Cycle through instruments
						for(int i=0; i<12; i++)
						{
							//First kill any notes being made by current instrument
							if(keyLocked[i])
							{
								midi.endNote(keyToPitch(key).toMidiFrequency(), instr);
								keyLocked[i] = false;
							}
						}
						instr = instr.next();
					}
					else if(key == 'R')
					{
						if(record)
						{
							record = false;
							setBackground(Color.black);
						}else{
							record = true;
							setBackground(Color.red);
							record_position = System.currentTimeMillis();
							recording = new LinkedList();
						}
					}
					else if(key == 'P')
					{
						if(!record)
						{
							setBackground(Color.green);
							while(!recording.isEmpty())
							{
								MusicAction action = (MusicAction)recording.get(0);
								if(action.note == 'w')
								{
									try
									{
										System.err.println(action.delay);
										Thread.sleep((int)action.delay);
									}catch(InterruptedException e2)
									{
										//do nothing
									}
									
								} else if (action.start){
									midi.beginNote(keyToPitch(action.note).toMidiFrequency(), instr);
								} else {
									midi.endNote(keyToPitch(action.note).toMidiFrequency(), instr);
								}
								recording.remove(0);
							}
							setBackground(Color.black);
						}
					}
				}
				return;
			}
		});

		addKeyListener(new KeyAdapter() {
			public void keyReleased(KeyEvent e) {
				char key = (char) e.getKeyCode();
				if(keyToInt(key) >= 0)
					keyLocked[keyToInt(key)] = false;
				if(keyToPitch(key) != null)
				{
					midi.endNote(keyToPitch(key).toMidiFrequency(), instr);
					if(record)
					{
						recording.add(new MusicAction('w', System.currentTimeMillis()-record_position, false));
						recording.add(new MusicAction(key, 0, false));
						record_position = System.currentTimeMillis();
					}					
				}
				return;
			}
		});
	}	

	private static Pitch keyToPitch(char key)
	{
		switch (key) {
		case '1':
			return new Pitch('C');
		case '2':
			return new Pitch('C').transpose(1);
		case '3':
	        return new Pitch('D');
		case '4':
	        return new Pitch('D').transpose(1);
		case '5':
	        return new Pitch('E');
		case '6':
	        return new Pitch('F');
		case '7':
	        return new Pitch('F').transpose(1);	
		case '8':
	        return new Pitch('G');
		case '9':
	        return new Pitch('G').transpose(1);
		case '0':
	        return new Pitch('A');
		case '-':
	        return new Pitch('A').transpose(1);			        
		case '=':
	        return new Pitch('B');			    
		}		
		return null;
	}
	
	private static int keyToInt(char key)
	{
		switch (key) {
		case '1':
			return 0;
		case '2':
			return 1;
		case '3':
	        return 2;
		case '4':
	        return 3;
		case '5':
	        return 4;
		case '6':
	        return 5;
		case '7':
	        return 6;	
		case '8':
	        return 7;
		case '9':
	        return 8;
		case '0':
	        return 9;
		case '-':
	        return 10;			        
		case '=':
	        return 11;	
		case 'I':
			return 12;
		case 'R':
			return 13;
		case 'P':
			return 14;
		}		
		return -1;
	}	
}