package cnf;

import cnf.Bool;
import cnf.Literal;
import cnf.Environment;

import java.util.LinkedList;

/**
 * Disjunctive clause of a CNF. All literals are disjuncts of the
 * form x ^ y ^ ~z etc.
 * @author scottyoung
 *
 */
public class Clause {
	
	public LinkedList <Literal>literals;
	
	/**
	 * Create a new, empty clause
	 */
	public Clause()
	{
		literals = new LinkedList<Literal>();
	}
	
	/**
	 * Adds a literal to the clause
	 * @param l - literal to be added
	 */
	public void add(Literal l)
	{
		literals.add(l);
	}
	
	/**
	 * Evaluate the clause, given an environment
	 * 
	 * Will return TRUE if there is at least one TRUE literal
	 * Will return FALSE only if there are ALL FALSE literals
	 * Will return UNDEFINED if there are no TRUE literals and at least one UNDEFINED literal
	 * 
	 * @param env - assignment of variables
	 * @return TRUE, FALSE or UNDEFINED
	 */
	public Bool eval(Environment env)
	{
		Bool val, defaultVal = Bool.FALSE;
		int index = 0;

		if(literals.size()==0)
			return(Bool.UNDEFINED);
		
		Literal l = literals.get(index);
		val = l.eval(env.eval(l.getID()));
		//if, at any point there's an undefined variable, the entire clause is undefined (unless it is true)
		if(val == Bool.UNDEFINED)
			defaultVal = Bool.UNDEFINED;
		
		while(index<literals.size()-1 && val!=Bool.TRUE)
		{
			index++;
			l = literals.get(index);
			val = l.eval(env.eval(l.getID()));
			if(val == Bool.UNDEFINED)
				defaultVal = Bool.UNDEFINED;
		}
		
		if(val == Bool.TRUE)
			return(Bool.TRUE);
		else
			return(defaultVal);
	}
	
	/**
	 * Returns the literal if the clause is atomic, otherwise returns null
	 */
	public Literal atomic(Environment env)
	{
		Literal atom = null;
		
		//A true clause cannot be atomic because it either has no unassigned variables, or
		//it has a true variable, in which case we can infer nothing about the remaining values.
		if(this.eval(env) == Bool.TRUE)
			return(null);
		
		for(int i=0; i<literals.size(); i++)
		{
			Literal l = literals.get(i);
			if(atom == null && env.eval(l.getID()) == Bool.UNDEFINED)
				atom = l;
			else if(atom != null && env.eval(l.getID()) == Bool.UNDEFINED)
				return(null);
		}
		return(atom);
	}
	
	public void print()
	{
		System.err.print("(");
		for(int i=0; i<literals.size(); i++)
		{
			literals.get(i).print();
			if(i < literals.size()-1)
				System.err.print(" V ");
		}		
		System.err.print(")");
	}
	
}
