package cnf;

import cnf.*;
import java.util.LinkedList;

public class Formula {

	private static long progressCount = 0;
	
	private LinkedList <Clause>clauses;
	
	/**
	 * Creates a new, blank formula
	 */
	public Formula()
	{
		clauses = new LinkedList<Clause>();

	}
	
	/**
	 * Adds a new clause to the CNF formula
	 */
	public void add(Clause c)
	{
		clauses.add(c);
	}
	
	/**
	 * Evaluates the formula, given an environment, and returns TRUE, FALSE or UNDEFINED
	 *
	 */
	public Bool eval(Environment env)
	{
		Bool val;
		int index = 0;

		if(clauses.size()==0)
			return(Bool.UNDEFINED);
		
		Clause c = clauses.get(index);
		val = c.eval(env);
		
		while(index<clauses.size()-1 && val==Bool.TRUE)
		{
			index++;
			c = clauses.get(index);
			val = c.eval(env);
		}
		
		return(val);
	}
	
	/**
	 * Solves the formula, determines whether or not it is satisfiable, and if so, returns
	 * an environment with all variables assigned TRUE or FALSE or returns null if the formula
	 * is unsatisfiable
	 * @param env - starting environment (may contain variables bound with UNDEFINED)
	 * @return either the empty environment or a valid assignment of variables for the given formula
	 */
	public Environment solve(Environment env)
	{

		//First, try unit propagation to eliminate possibilities
		Environment solution = this.unitPropagate(env);
		
		//TEMPORARY
		//System.err.println("---NEW ITERATION---");
		//solution.display();
		progress();
		
		Bool eval = this.eval(solution);
		//First check if the given environment is a solution
		if(eval == Bool.TRUE)
			return(solution);
		//If it's FALSE then no variables can be assigned, so solve fails
		else if(eval == Bool.FALSE)
			return(null);
		
		//Pick a variable to assign
		int choice = solution.choose();
		//If there are no unassigned variables
		if(choice==-1)
			return(null);
		
		Environment solutionTrue = new Environment(solution);
		Environment solutionFalse = new Environment(solution);
		
		//First, assign it to be true
		solutionTrue.add(choice, Bool.TRUE);
		solutionTrue = this.solve(solutionTrue);
		if(solutionTrue != null)
			return(solutionTrue);
		
		//Now try assigning it false
		solutionFalse.add(choice, Bool.FALSE);
		solutionFalse = this.solve(solutionFalse);
		if(solutionFalse==null)
			return(null);
		else
			return(solutionFalse);
	}
	
	
	
	/**
	 * Unit propagates until either (a) all clauses contain at least 2 UNDEFINED literals or
	 * all variables have been defined (i.e. choose() returns -1)
	 * @param env - starting environment
	 * @return environment with possibly extra assignments to TRUE or FALSE of undefined variables
	 */
	public Environment unitPropagate(Environment env)
	{
		Environment prop = new Environment(env);
		
		//Go through every clause
		for(int i = 0; i < clauses.size(); i++)
		{
			Clause c = clauses.get(i);
			Literal atom = c.atomic(prop);
			if(atom!=null)
			{
				if(atom.getNeg())
					prop.add(atom.getID(), Bool.FALSE);
				else
					prop.add(atom.getID(), Bool.TRUE);
			}
			
		}
		return(prop);
	}
	
	/**
	 * Prints the contents of the formula in clause ^ clause format
	 */
	public void print()
	{
		for(int i=0; i<clauses.size(); i++)
		{
			clauses.get(i).print();
			if(i < clauses.size()-1)
				System.err.println(" ^ ");
		}
	
	}
	
	public static void progress()
	{
		if(progressCount%10 == 0)
		{
			System.err.print('.');
		}
		if(progressCount%200 == 0)
		{
			System.err.print('\n');
			progressCount = 0;
		}
		progressCount++;
		
	}
	
}
