package client;

import java.awt.Container;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.io.PrintWriter;
import java.util.ArrayList;
import java.util.List;

import javax.swing.DefaultListModel;
import javax.swing.JButton;
import javax.swing.JFrame;
import javax.swing.JLabel;
import javax.swing.JList;
import javax.swing.JMenu;
import javax.swing.JMenuBar;
import javax.swing.JMenuItem;
import javax.swing.JOptionPane;
import javax.swing.JScrollPane;
import javax.swing.ListSelectionModel;
import javax.swing.SwingUtilities;

import controller.Command;
import controller.Protocol;
import controller.Tuple;


import layout.TableLayout;

/**
 * The HubPane is the main window which controls the room and user list
 * 
 */
public class HubPane extends JFrame {
	
	//---------------GUI ELEMENTS--------------------
	//Options menu bar
	private final JMenuBar menu;
	private final JMenu menuOptions;
	private final JMenuItem menuChangeAlias;
	private final JMenuItem menuCreateRoom;
	private final JMenuItem menuExit;
	
	//Text fields
	private final JLabel displayAlias;
	private final JLabel displayRooms;
	private final JLabel displayUsers;
	
	//Buttons
	private final JButton changeAlias;
	private final JButton createRoom;
	private final JButton joinRoom;
	
	//Chatroom list
	private final JScrollPane paneRooms;
	private final JList listRooms;
	private final DefaultListModel listRoomsModel;
	
	//Users list
	private final JScrollPane paneUsers;
	private final JList listUsers;
	private final DefaultListModel listUsersModel;
	//------------------------------------------------
	
	//----------DATA ELEMENTS-------------------------
	
	List<Tuple> roomList;
	List<String> userList;
	String myAlias;
	List<ChatroomPane> chatWindows;
	
	PrintWriter writer;
	
	//------------------------------------------------
	
	public HubPane(PrintWriter theWriter){
		super("Incognito Chat v1.0");
		setDefaultCloseOperation(EXIT_ON_CLOSE);
		
		chatWindows = new ArrayList<ChatroomPane>();
		writer = theWriter;
		
		//Set the layout of the hub
		double size[][] = {{5, TableLayout.PREFERRED, 5, TableLayout.PREFERRED, TableLayout.PREFERRED, TableLayout.FILL, 5, TableLayout.PREFERRED, 5},
						   {5, TableLayout.PREFERRED, 5, TableLayout.PREFERRED, TableLayout.FILL, 5, TableLayout.PREFERRED, 5}};
		this.setLayout(new TableLayout(size));
		
		//Add the elements
		this.add(displayAlias = new JLabel("Alias: ???"), "1, 1, 3, 1");
		this.add(displayRooms = new JLabel("Chatrooms Available (?):"), "1, 3, 3, 1");
		this.add(displayUsers = new JLabel("Users Online (?):"), "7, 3, 1, 1");
		
		this.add(changeAlias = new JButton("Change Alias..."), "4, 1, 1, 1");
		this.add(createRoom = new JButton("Create Chatroom..."), "1, 6, 1, 1");
		this.add(joinRoom = new JButton("Join Chatroom"), "3, 6, 1, 1");
		
		listRoomsModel = new DefaultListModel();
		listRooms = new JList(listRoomsModel);
		listRooms.setSelectionMode(ListSelectionModel.SINGLE_SELECTION);
		this.add(paneRooms = new JScrollPane(listRooms), "1, 4, 5, 1");

		listUsersModel = new DefaultListModel();
		listUsers = new JList(listUsersModel);
		this.add(paneUsers = new JScrollPane(listUsers), "7, 4, 1, 3");
		
		//Initialize the menu
		this.setJMenuBar(menu = new JMenuBar());
		menu.add(menuOptions = new JMenu("Options"));
		menuOptions.add(menuChangeAlias = new JMenuItem("Change Alias..."));
		menuOptions.add(menuCreateRoom = new JMenuItem("Create Room..."));
		menuOptions.add(menuExit = new JMenuItem("Exit"));
		
		this.pack();
		
		//Add action listeners
		
		changeAlias.addActionListener(new ActionListener(){
        	public void actionPerformed(ActionEvent e)
        	{
        		changeAliasDialog();
        	}
        });		
		
		createRoom.addActionListener(new ActionListener(){
        	public void actionPerformed(ActionEvent e)
        	{
        		createRoomDialog();
        	}
        });	
		
		joinRoom.addActionListener(new ActionListener(){
        	public void actionPerformed(ActionEvent e)
        	{
        		joinRoom();
        	}
        });		
		
		menuChangeAlias.addActionListener(new ActionListener(){
        	public void actionPerformed(ActionEvent e)
        	{
        		changeAliasDialog();
        	}
        });			
		
		menuCreateRoom.addActionListener(new ActionListener(){
        	public void actionPerformed(ActionEvent e)
        	{
        		createRoomDialog();
        	}
        });			
		
		menuExit.addActionListener(new ActionListener(){
        	public void actionPerformed(ActionEvent e)
        	{
        		System.exit(0);
        	}
        });			
		
	}
	
	//---------------------------ACTIONS-----------------------------
	/**
	 * Prompts user to enter a new chatroom name, then requests the
	 * creation from the server and closes.
	 */
	public void createRoomDialog()
	{
		String roomname = promptForInput("New chatroom name:", "Create Chatroom");
		send(new Protocol(Command.CREATE, roomname));
	}
	
	/**
	 * Prompts the user to enter a new alias, then requests the alias
	 * change from the server and closes
	 */
	public void changeAliasDialog()
	{
		String alias = promptForInput("New alias:", "Change Alias");
		send(new Protocol(Command.ALIAS, alias));
	}
	
	/**
	 * Finds the selected element of the paneRooms, finds the roomname
	 * from the roomList and sends a request to join the room
	 */
	public void joinRoom()
	{
		if(listRooms.getSelectedValue() != null)
		{
			RoomString rawTag = (RoomString)listRooms.getSelectedValue();
			send(new Protocol(Command.JOIN, rawTag.roomname));
		}
	}
	
	private String promptForInput(String message, String title) {
		return (String)
		  JOptionPane.showInputDialog(
				paneRooms, 
				message, 
				title,
				JOptionPane.PLAIN_MESSAGE, 
				null, 
				null, 
				"");		
	}	
	
	//---------------------------------------------------------------
	
	/**
	 * Processes a message from the server to update
	 * @param update
	 */
	public void update(final Protocol update)
	{
		switch(update.command)
		{
		//We've received confirmation that we've joined a chatroom, open a new ChatroomPane with that name
		//and start that process
		case INVITED:
			SwingUtilities.invokeLater(new Runnable() {
	            public void run () {
	                ChatroomPane newRoom = new ChatroomPane(update.nameField, writer);
	                newRoom.setVisible(true);
	                chatWindows.add(newRoom);
	        		//Now query the server for the transcript since if we were delayed in setup, the
	        		//transcript may have already been sent and discarded
	        		writer.println("[TRANSCRIPT]["+update.nameField+"][none]");
	        		writer.println("[ROOMLIST][none`0]");
	        		writer.flush();	                
	            }
	        });
			break;
		
			
		//We've been assigned the alias, so update our alias and all labels which display it
		case ASSIGNED:
			SwingUtilities.invokeLater(new Runnable() {
				public void run(){
					myAlias = update.nameField;
					displayAlias.setText("Alias: "+myAlias);
				}
			});
			break;
		
			
		//We've been given an updated transcript, clear out the old transcript and replace it with the new one
		case TRANSCRIPT:
			SwingUtilities.invokeLater(new Runnable() {
				public void run(){
					ChatroomPane theRoom = null;
					cleanChatWindows();
					for(ChatroomPane c : chatWindows)
					{
						if(c.roomname.equals(update.nameField))
						{
							theRoom = c;
							break;
						}
					}
					//Only update if the window is active (may be false in normal operation if window takes too long to load)
					if(theRoom!=null)
						theRoom.update(update);
				}
			});
			break;
		
		//We've been given an updated roomlist, change the current roomlist and update the display panel
		case ROOMLIST:
			SwingUtilities.invokeLater(new Runnable() {
				public void run(){
					roomList = update.roomList;
					updateRoomListDisplay();
					for(ChatroomPane c : chatWindows)
					{
						c.update(update);
					}
				}
			});
			break;
		
			
		//We've been given an updated userlist, change the current userlist and update the display panel
		case USERLIST:
			SwingUtilities.invokeLater(new Runnable() {
				public void run(){
					userList = update.userList;
					updateUserListDisplay();
				}
			});
			break;
		
		
		default:
			System.err.println("Warning, client cannot parse command, "+update.command);
			break;
		}
	}
	
	/**
	 * Removes all invisible windows from chatWindows (closed)
	 */
	public void cleanChatWindows()
	{
		List<ChatroomPane> toRemove = new ArrayList<ChatroomPane>();
		for(ChatroomPane c : chatWindows)
			if(!c.isVisible())
				toRemove.add(c);
		for(ChatroomPane c : toRemove)
			chatWindows.remove(c);
	}
	
	/**
	 * Updates the list model for the room display
	 */
	public void updateRoomListDisplay()
	{
		displayRooms.setText("Chatrooms Available ("+roomList.size()+"):");
		listRoomsModel.clear();
		for(Tuple t : roomList)
			listRoomsModel.addElement(new RoomString(t)); 
	}
	
	public void updateUserListDisplay()
	{
		displayUsers.setText("Users Online ("+userList.size()+"):");
		listUsersModel.clear();
		for(String s : userList)
			listUsersModel.addElement(s);
	}
	
	/**
	 * Sends a protocol message to the server
	 */
	public void send(Protocol p)
	{
		writer.print("["+p.command+"]");
		switch(p.command)
		{
		case USERLIST:
			for(String s : p.userList)
				writer.print("["+s+"]");
			break;
		case ROOMLIST:
			for(Tuple t : p.roomList)
				writer.print("["+t.name+"`"+t.number+"]");
			break;
		case MSG: case TRANSCRIPT:
			writer.print("["+p.nameField+"]["+p.textField+"]");
			break;
		default:
			writer.print("["+p.nameField+"]");
		}
		writer.print("\n");
		writer.flush();
	}
	
	class RoomString {
		public String roomname;
		public String display;
		public RoomString(Tuple t)
		{
			roomname = t.name;
			display = t.name + " (" + t.number + ")";
		}
		@Override
		public String toString(){
			return display;
		}
	}	
	
	//For viewing the hub pane (doesn't do anything)
	public static void main(String[] args)
	{
		SwingUtilities.invokeLater(new Runnable() {
            public void run () {
                // Make and display the HubPane window.
                new HubPane(null).setVisible(true);
            }
        });
	}

}
